from typing import Any, Optional, Set, Type

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.registry.command_registry import CommandRegistry
from bus_station.passengers.passenger_class_resolver import PassengerClassResolver
from bus_station.passengers.passenger_record.in_memory_passenger_record_repository import (
    InMemoryPassengerRecordRepository,
)
from bus_station.passengers.passenger_record.passenger_record import PassengerRecord
from bus_station.shared_terminal.bus_stop_resolver.bus_stop_resolver import BusStopResolver
from bus_station.shared_terminal.fqn_getter import FQNGetter


class InMemoryCommandRegistry(CommandRegistry):
    def __init__(
        self,
        in_memory_repository: InMemoryPassengerRecordRepository,
        fqn_getter: FQNGetter,
        command_handler_resolver: BusStopResolver[CommandHandler],
        passenger_class_resolver: PassengerClassResolver,
    ):
        self.__in_memory_repository = in_memory_repository
        self.__fqn_getter = fqn_getter
        self.__command_handler_resolver = command_handler_resolver
        self.__passenger_class_resolver = passenger_class_resolver

    def _register(self, command: Type[Command], handler: CommandHandler, handler_contact: Any) -> None:
        self.__in_memory_repository.save(
            PassengerRecord(
                passenger_name=command.__name__,
                passenger_fqn=self.__fqn_getter.get(command),
                destination_fqn=self.__fqn_getter.get(handler),
                destination_contact=handler_contact,
            )
        )

    def get_command_destination(self, command: Type[Command]) -> Optional[CommandHandler]:
        command_records = self.__in_memory_repository.find_by_passenger_name(command.__name__)
        if command_records is None:
            return None

        command_handler_fqn = command_records[0].destination_fqn
        return self.__command_handler_resolver.resolve_from_fqn(command_handler_fqn)

    def get_command_destination_contact(self, command: Type[Command]) -> Optional[Any]:
        command_records = self.__in_memory_repository.find_by_passenger_name(command.__name__)
        if command_records is None:
            return None

        return command_records[0].destination_contact

    def get_commands_registered(self) -> Set[Type[Command]]:
        commands_registered = set()
        for command_record in self.__in_memory_repository.all():
            command = self.__passenger_class_resolver.resolve_from_fqn(command_record.passenger_fqn)
            if command is None or not issubclass(command, Command):
                continue
            commands_registered.add(command)
        return commands_registered

    def unregister(self, command: Type[Command]) -> None:
        self.__in_memory_repository.delete_by_passenger_name(command.__name__)



from bus_station.shared_terminal.factories.proxy_class_factory import ProxyClassFactory
from bus_station.tracking_terminal.models.command_tracking import CommandTracking
from bus_station.tracking_terminal.models.event_tracking import EventTracking
from bus_station.tracking_terminal.models.query_tracking import QueryTracking

SACommandTrackingProxy = ProxyClassFactory.get_proxy(CommandTracking)
SAEventTrackingProxy = ProxyClassFactory.get_proxy(EventTracking)
SAQueryTrackingProxy = ProxyClassFactory.get_proxy(QueryTracking)



import os
import signal
from multiprocessing import Process
from typing import ClassVar, Optional

from rpyc import Connection, connect

from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer
from bus_station.query_terminal.bus.query_bus import QueryBus
from bus_station.query_terminal.handler_not_found_for_query import HandlerNotFoundForQuery
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse
from bus_station.query_terminal.registry.remote_query_registry import RemoteQueryRegistry
from bus_station.query_terminal.rpyc_query_service import RPyCQueryService
from bus_station.query_terminal.serialization.query_response_deserializer import QueryResponseDeserializer
from bus_station.query_terminal.serialization.query_response_serializer import QueryResponseSerializer
from bus_station.shared_terminal.rpyc_server import RPyCServer
from bus_station.shared_terminal.runnable import Runnable


class RPyCQueryBus(QueryBus, Runnable):
    __SELF_ADDR_PATTERN: ClassVar[str] = "{host}:{port}"

    def __init__(
        self,
        self_host: str,
        self_port: int,
        query_serializer: PassengerSerializer,
        query_deserializer: PassengerDeserializer,
        query_response_serializer: QueryResponseSerializer,
        query_response_deserializer: QueryResponseDeserializer,
        query_registry: RemoteQueryRegistry,
        query_receiver: PassengerReceiver[Query, QueryHandler],
    ):
        QueryBus.__init__(self, query_receiver)
        Runnable.__init__(self)
        self.__self_host = self_host
        self.__self_port = self_port
        self.__query_serializer = query_serializer
        self.__query_deserializer = query_deserializer
        self.__query_response_serializer = query_response_serializer
        self.__query_response_deserializer = query_response_deserializer
        self.__query_registry = query_registry
        self.__rpyc_service = RPyCQueryService(
            self.__query_deserializer, self.__query_response_serializer, self._query_receiver
        )
        self.__rpyc_server: Optional[RPyCServer] = None
        self.__server_process: Optional[Process] = None

    def _start(self):
        for query in self.__query_registry.get_queries_registered():
            handler = self.__query_registry.get_query_destination(query)
            self.__rpyc_service.register(query, handler)

        self.__rpyc_server = RPyCServer(
            rpyc_service=self.__rpyc_service,
            port=self.__self_port,
        )
        self.__server_process = Process(target=self.__rpyc_server.run)
        self.__server_process.start()

    def transport(self, passenger: Query) -> QueryResponse:
        query_handler_addr = self.__query_registry.get_query_destination_contact(passenger.__class__)
        if query_handler_addr is None:
            raise HandlerNotFoundForQuery(passenger.__class__.__name__)

        rpyc_client = self.__get_rpyc_client(query_handler_addr)
        query_response = self.__execute_query(rpyc_client, passenger)
        rpyc_client.close()
        return query_response

    def __get_rpyc_client(self, handler_addr: str) -> Connection:
        host, port = handler_addr.split(":")
        return connect(host, port=port)

    def __execute_query(self, client: Connection, query: Query) -> QueryResponse:
        serialized_query = self.__query_serializer.serialize(query)
        serialized_query_response = getattr(client.root, query.__class__.__name__)(serialized_query)
        return self.__query_response_deserializer.deserialize(serialized_query_response)

    def _stop(self) -> None:
        server_process = self.__server_process
        if server_process is not None and server_process.pid is not None:
            os.kill(server_process.pid, signal.SIGINT)
            server_process.join()



from ctypes import c_bool
from multiprocessing import Queue, Value
from queue import Empty

from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.shared_terminal.bus_stop import BusStop


class ProcessPassengerWorker:
    def __init__(
        self,
        queue: Queue,
        passenger_bus_stop: BusStop,
        passenger_receiver: PassengerReceiver,
        passenger_deserializer: PassengerDeserializer,
    ):
        self.__queue = queue
        self.__passenger_bus_stop = passenger_bus_stop
        self.__passenger_deserializer = passenger_deserializer
        self.__passenger_receiver = passenger_receiver
        self.__running = Value(c_bool, False)

    def work(self) -> None:
        self.__running.value = True
        while self.__running.value:
            try:
                item = self.__queue.get(timeout=1.0)
            except Empty:
                continue
            except KeyboardInterrupt:
                self.__running.value = False
                continue
            self.__handle_item(item)

    def __handle_item(self, item: str) -> None:
        deserialized_item = self.__passenger_deserializer.deserialize(item)
        self.__passenger_receiver.receive(deserialized_item, self.__passenger_bus_stop)

    def stop(self):
        self.__running.value = False



from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.middleware.event_middleware import EventMiddleware
from bus_station.passengers.reception.passenger_middleware_receiver import PassengerMiddlewareReceiver


class EventMiddlewareReceiver(PassengerMiddlewareReceiver[Event, EventConsumer, EventMiddleware]):
    def receive(self, passenger: Event, passenger_bus_stop: EventConsumer) -> None:
        middlewares = list(self._get_middlewares())
        for middleware in middlewares:
            middleware.before_consume(passenger, passenger_bus_stop)

        passenger_bus_stop.consume(passenger)

        for middleware in reversed(middlewares):
            middleware.after_consume(passenger, passenger_bus_stop)



from abc import abstractmethod

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.passengers.passenger_middleware import PassengerMiddleware


class CommandMiddleware(PassengerMiddleware):
    @abstractmethod
    def before_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        pass

    @abstractmethod
    def after_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        pass



from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar, Generic, TypeVar

C = TypeVar("C")


@dataclass(frozen=True)
class PassengerRecord(Generic[C]):
    __STR_SEPARATOR: ClassVar[str] = "-"
    __STR_PATTERN: ClassVar[
        str
    ] = "{passenger_name}{separator}{passenger_fqn}{separator}{destination_fqn}{separator}{destination_contact}"

    passenger_name: str
    passenger_fqn: str
    destination_fqn: str
    destination_contact: C

    @classmethod
    def from_str(cls, passenger_record_str: str) -> PassengerRecord[str]:
        passenger_record_attrs = passenger_record_str.split(sep=cls.__STR_SEPARATOR)
        if len(passenger_record_attrs) > 4:
            raise ValueError(
                f"Passenger record string {passenger_record_str} is not a valid representation of PassengerRecord"
            )

        return cls[str](
            passenger_name=passenger_record_attrs[0],
            passenger_fqn=passenger_record_attrs[1],
            destination_fqn=passenger_record_attrs[2],
            destination_contact=passenger_record_attrs[3],
        )

    def __str__(self) -> str:
        destination_contact = self.destination_contact
        if not isinstance(destination_contact, str):
            destination_contact = str(destination_contact)
        return self.__STR_PATTERN.format(
            passenger_name=self.passenger_name,
            separator=self.__STR_SEPARATOR,
            passenger_fqn=self.passenger_fqn,
            destination_fqn=self.destination_fqn,
            destination_contact=destination_contact,
        )



from bus_station.event_terminal.bus.event_bus import EventBus
from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.registry.in_memory_event_registry import InMemoryEventRegistry
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver


class SyncEventBus(EventBus):
    def __init__(self, event_registry: InMemoryEventRegistry, event_receiver: PassengerReceiver[Event, EventConsumer]):
        super().__init__(event_receiver)
        self.__event_registry = event_registry

    def transport(self, passenger: Event) -> None:
        event_consumers = self.__event_registry.get_event_destination_contacts(passenger.__class__)
        if event_consumers is None:
            return

        for event_consumer in event_consumers:
            self._event_receiver.receive(passenger, event_consumer)



from abc import abstractmethod

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.passengers.passenger_middleware import PassengerMiddleware


class EventMiddleware(PassengerMiddleware):
    @abstractmethod
    def before_consume(self, passenger: Event, bus_stop: EventConsumer) -> None:
        pass

    @abstractmethod
    def after_consume(self, passenger: Event, bus_stop: EventConsumer) -> None:
        pass



from threading import Thread

from bus_station.command_terminal.bus.command_bus import CommandBus
from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.in_memory_command_registry import InMemoryCommandRegistry
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver


class ThreadedCommandBus(CommandBus):
    def __init__(
        self, command_registry: InMemoryCommandRegistry, command_receiver: PassengerReceiver[Command, CommandHandler]
    ):
        super().__init__(command_receiver)
        self.__command_registry = command_registry

    def transport(self, passenger: Command) -> None:
        command_handler = self.__command_registry.get_command_destination_contact(passenger.__class__)
        if command_handler is None:
            raise HandlerNotFoundForCommand(passenger.__class__.__name__)

        execution_thread = Thread(target=self._command_receiver.receive, args=(passenger, command_handler))
        execution_thread.start()



class HandlerNotFoundForQuery(Exception):
    def __init__(self, query_name: str):
        self.query_name = query_name
        super(HandlerNotFoundForQuery, self).__init__(f"Handler for query {query_name} not found")



from logging import Logger

from bus_station.query_terminal.middleware.query_middleware import QueryMiddleware
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse


class LoggingQueryMiddleware(QueryMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger

    def before_handle(self, passenger: Query, bus_stop: QueryHandler) -> None:
        self.__logger.info(f"Starting handling query {passenger} with {bus_stop.__class__.__name__}")

    def after_handle(self, passenger: Query, bus_stop: QueryHandler, query_response: QueryResponse) -> QueryResponse:
        self.__logger.info(
            f"Finished handling query {passenger} with {bus_stop.__class__.__name__} with response: {query_response}"
        )
        return query_response



from dataclasses import asdict
from datetime import datetime
from typing import Any, Type

from bus_station.passengers.passenger import Passenger
from bus_station.shared_terminal.bus_stop import BusStop
from bus_station.tracking_terminal.models.passenger_tracking import PassengerTracking
from bus_station.tracking_terminal.repositories.passenger_tracking_repository import PassengerTrackingRepository
from bus_station.tracking_terminal.trackers.passenger_tracking_not_found import PassengerTrackingNotFound


class PassengerTracker:
    def __init__(self, passenger_tracking_repository: PassengerTrackingRepository):
        self.__repository = passenger_tracking_repository

    def _get_tracking_model(self, passenger: Passenger) -> Type:
        return PassengerTracking

    def start_tracking(self, passenger: Passenger, bus_stop: BusStop) -> None:
        tracking_model = self._get_tracking_model(passenger)
        tracking = tracking_model(
            id=str(id(passenger)),
            name=passenger.__class__.__name__,
            executor_name=bus_stop.__class__.__name__,
            data=asdict(passenger),
            execution_start=datetime.now(),
            execution_end=None,
        )
        self.__repository.save(tracking)

    def end_tracking(self, passenger: Passenger, **track_data: Any) -> None:
        passenger_id = str(id(passenger))
        tracking = self.__repository.find_by_id(passenger_id)
        if tracking is None:
            raise PassengerTrackingNotFound(passenger.__class__.__name__, passenger_id)
        tracking.execution_end = datetime.now()
        self.__set_track_data(tracking, **track_data)
        self.__repository.save(tracking)

    def __set_track_data(self, passenger_tracking: PassengerTracking, **track_data: Any) -> None:
        for track_data_key, track_data_value in track_data.items():
            setattr(passenger_tracking, track_data_key, track_data_value)



from abc import abstractmethod
from typing import Optional, Protocol, TypeVar

from bus_station.shared_terminal.bus_stop import BusStop

S = TypeVar("S", bound=BusStop)


class BusStopResolver(Protocol[S]):
    @abstractmethod
    def resolve_from_fqn(self, bus_stop_fqn: str) -> Optional[S]:
        pass



import os
import signal
from multiprocessing import Process
from typing import ClassVar, Optional

import requests
from jsonrpcclient import request
from jsonrpcclient.responses import Error, to_result

from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer
from bus_station.query_terminal.bus.query_bus import QueryBus
from bus_station.query_terminal.handler_not_found_for_query import HandlerNotFoundForQuery
from bus_station.query_terminal.json_rpc_query_server import JsonRPCQueryServer
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_execution_failed import QueryExecutionFailed
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse
from bus_station.query_terminal.registry.remote_query_registry import RemoteQueryRegistry
from bus_station.query_terminal.serialization.query_response_deserializer import QueryResponseDeserializer
from bus_station.query_terminal.serialization.query_response_serializer import QueryResponseSerializer
from bus_station.shared_terminal.runnable import Runnable


class JsonRPCQueryBus(QueryBus, Runnable):
    __SELF_ADDR_PATTERN: ClassVar[str] = "http://{host}:{port}/"

    def __init__(
        self,
        self_host: str,
        self_port: int,
        query_serializer: PassengerSerializer,
        query_deserializer: PassengerDeserializer,
        query_response_serializer: QueryResponseSerializer,
        query_response_deserializer: QueryResponseDeserializer,
        query_registry: RemoteQueryRegistry,
        query_receiver: PassengerReceiver[Query, QueryHandler],
    ):
        QueryBus.__init__(self, query_receiver)
        Runnable.__init__(self)
        self.__self_host = self_host
        self.__self_port = self_port
        self.__query_serializer = query_serializer
        self.__query_deserializer = query_deserializer
        self.__query_registry = query_registry
        self.__query_response_serializer = query_response_serializer
        self.__query_response_deserializer = query_response_deserializer
        self.__json_rpc_query_server = JsonRPCQueryServer(
            passenger_deserializer=self.__query_deserializer,
            passenger_receiver=self._query_receiver,
            query_response_serializer=self.__query_response_serializer,
        )
        self.__server_process: Optional[Process] = None

    def _start(self):
        for query in self.__query_registry.get_queries_registered():
            handler = self.__query_registry.get_query_destination(query)
            self.__json_rpc_query_server.register(query, handler)

        self.__server_process = Process(target=self.__json_rpc_query_server.run, args=(self.__self_port,))
        self.__server_process.start()

    def transport(self, passenger: Query) -> QueryResponse:
        handler_address = self.__query_registry.get_query_destination_contact(passenger.__class__)
        if handler_address is None:
            raise HandlerNotFoundForQuery(passenger.__class__.__name__)

        return self.__execute_query(passenger, handler_address)

    def __execute_query(self, query: Query, query_handler_addr: str) -> QueryResponse:
        serialized_query = self.__query_serializer.serialize(query)

        request_response = requests.post(
            query_handler_addr, json=request(query.__class__.__name__, params=(serialized_query,))
        )
        json_rpc_response = to_result(request_response.json())

        if isinstance(json_rpc_response, Error):
            raise QueryExecutionFailed(query, json_rpc_response.message)
        return self.__query_response_deserializer.deserialize(json_rpc_response.result)

    def _stop(self) -> None:
        server_process = self.__server_process
        if server_process is not None and server_process.pid is not None:
            os.kill(server_process.pid, signal.SIGINT)
            server_process.join()



from abc import abstractmethod

from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse
from bus_station.shared_terminal.bus import Bus


class QueryBus(Bus[Query]):
    def __init__(self, query_receiver: PassengerReceiver[Query, QueryHandler]):
        self._query_receiver = query_receiver

    @abstractmethod
    def transport(self, passenger: Query) -> QueryResponse:
        pass



from typing import Optional

from sqlalchemy.orm import Session

from bus_station.tracking_terminal.models.proxy_definitions import SAQueryTrackingProxy
from bus_station.tracking_terminal.models.query_tracking import QueryTracking
from bus_station.tracking_terminal.passenger_tracking_to_proxy_transformer import PassengerTrackingToProxyTransformer
from bus_station.tracking_terminal.proxy_to_passenger_tracking_transformer import ProxyToPassengerTrackingTransformer
from bus_station.tracking_terminal.repositories.query_tracking_repository import QueryTrackingRepository


class SQLAlchemyQueryTrackingRepository(QueryTrackingRepository):
    def __init__(
        self,
        sqlalchemy_session: Session,
        passenger_tracking_transformer: PassengerTrackingToProxyTransformer,
        proxy_transformer: ProxyToPassengerTrackingTransformer,
    ):
        self.__session = sqlalchemy_session
        self.__passenger_tracking_transformer = passenger_tracking_transformer
        self.__proxy_transformer = proxy_transformer

    def save(self, passenger_tracking: QueryTracking) -> None:
        proxy_instance = self.__passenger_tracking_transformer.transform(passenger_tracking, SAQueryTrackingProxy)
        self.__session.merge(proxy_instance)
        self.__session.commit()

    def find_by_id(self, passenger_tracking_id: str) -> Optional[QueryTracking]:
        proxy_instance = self.__session.query(SAQueryTrackingProxy).filter_by(id=passenger_tracking_id).one_or_none()
        if proxy_instance is None:
            return None
        return self.__proxy_transformer.transform(proxy_instance, QueryTracking)



from abc import ABC, abstractmethod
from typing import Iterable, Optional, Type

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.registry.event_registry import EventRegistry


class RemoteEventRegistry(EventRegistry, ABC):
    @abstractmethod
    def _register(self, event: Type[Event], consumer: EventConsumer, consumer_contact: str) -> None:
        pass

    @abstractmethod
    def get_event_destination_contacts(self, event: Type[Event]) -> Optional[Iterable[str]]:
        pass



import time
from logging import Logger

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.middleware.event_middleware import EventMiddleware


class TimingEventMiddleware(EventMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger
        self.__start_time = None

    def before_consume(self, passenger: Event, bus_stop: EventConsumer) -> None:
        self.__start_time = time.time()

    def after_consume(self, passenger: Event, bus_stop: EventConsumer) -> None:
        execution_time = time.time() - self.__start_time
        self.__logger.info(f"Event {passenger} consumed by {bus_stop.__class__.__name__} in {execution_time} seconds")



from abc import abstractmethod
from typing import Protocol

from bus_station.query_terminal.query_response import QueryResponse


class QueryResponseSerializer(Protocol):
    @abstractmethod
    def serialize(self, query_response: QueryResponse) -> str:
        pass



from abc import ABC, abstractmethod
from typing import Type

from sqlalchemy import MetaData, Table
from sqlalchemy.orm import mapper


class SQLAlchemyMapper(ABC):
    def __init__(self, db_metadata: MetaData):
        self._db_metadata = db_metadata

    @property
    @abstractmethod
    def model(self) -> Type:
        pass

    @property
    @abstractmethod
    def _proxy(self) -> Type:
        pass

    @property
    @abstractmethod
    def table(self) -> Table:
        pass

    def map(self) -> None:
        mapper(self._proxy, self.table)



from abc import abstractmethod
from typing import Any, Optional, Protocol, TypeVar

from bus_station.passengers.passenger import Passenger
from bus_station.shared_terminal.bus_stop import BusStop

S = TypeVar("S", bound=BusStop)
P = TypeVar("P", bound=Passenger)


class PassengerReceiver(Protocol[P, S]):
    @abstractmethod
    def receive(self, passenger: P, passenger_bus_stop: S) -> Optional[Any]:
        pass



from logging import Logger

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.middleware.event_middleware import EventMiddleware


class LoggingEventMiddleware(EventMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger

    def before_consume(self, passenger: Event, bus_stop: EventConsumer) -> None:
        self.__logger.info(f"Starting consuming event {passenger} with {bus_stop.__class__.__name__}")

    def after_consume(self, passenger: Event, bus_stop: EventConsumer) -> None:
        self.__logger.info(f"Finished consuming event {passenger} with {bus_stop.__class__.__name__}")



from typing import Type

from sqlalchemy import JSON, Column, DateTime, String, Table

from bus_station.tracking_terminal.mappers.sqlalchemy.sqlalchemy_mapper import SQLAlchemyMapper
from bus_station.tracking_terminal.models.proxy_definitions import SAQueryTrackingProxy
from bus_station.tracking_terminal.models.query_tracking import QueryTracking


class SQLAlchemyQueryTrackingMapper(SQLAlchemyMapper):
    @property
    def table(self) -> Table:
        return Table(
            "query_tracking",
            self._db_metadata,
            Column("id", String(255), primary_key=True),
            Column("name", String(255), nullable=False),
            Column("executor_name", String(255), nullable=False),
            Column("data", JSON, nullable=False),
            Column("response_data", JSON, nullable=True),
            Column("execution_start", DateTime, nullable=False),
            Column("execution_end", DateTime, nullable=True),
        )

    @property
    def model(self) -> Type:
        return QueryTracking

    @property
    def _proxy(self) -> Type:
        return SAQueryTrackingProxy



from abc import abstractmethod

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.shared_terminal.bus import Bus


class CommandBus(Bus[Command]):
    def __init__(self, command_receiver: PassengerReceiver[Command, CommandHandler]):
        self._command_receiver = command_receiver

    @abstractmethod
    def transport(self, passenger: Command) -> None:
        pass



from typing import Any, Dict, Iterable, List, Optional

from bus_station.passengers.passenger_record.passenger_record import PassengerRecord
from bus_station.passengers.passenger_record.passenger_record_repository import PassengerRecordRepository


class InMemoryPassengerRecordRepository(PassengerRecordRepository):
    def __init__(self):
        self.__passenger_records: Dict[str, List[PassengerRecord[Any]]] = {}

    def save(self, passenger_record: PassengerRecord[Any]) -> None:
        if passenger_record.passenger_name in self.__passenger_records:
            self.__passenger_records[passenger_record.passenger_name].append(passenger_record)
            return
        self.__passenger_records[passenger_record.passenger_name] = [passenger_record]

    def find_by_passenger_name(self, passenger_name: str) -> Optional[List[PassengerRecord]]:
        return self.__passenger_records.get(passenger_name)

    def all(self) -> Iterable[PassengerRecord[Any]]:
        for passenger_records in self.__passenger_records.values():
            for passenger_record in passenger_records:
                yield passenger_record

    def delete_by_passenger_name(self, passenger_name: str) -> None:
        if passenger_name in self.__passenger_records:
            del self.__passenger_records[passenger_name]



from abc import ABC, abstractmethod
from functools import wraps

from bus_station.shared_terminal.runnable_already_runnning_exception import RunnableAlreadyRunningException
from bus_station.shared_terminal.runnable_not_running_exception import RunnableNotRunningException


def is_running(func):
    @wraps(func)
    def checked_running(*args, **kwargs):
        if len(args) == 0 or not isinstance(args[0], Runnable):
            raise ValueError("is_running only applicable to Runnable instances")
        runnable = args[0]
        if runnable.running is False:
            raise RunnableNotRunningException(runnable.__class__.__name__)
        return func(*args, **kwargs)

    return checked_running


def is_not_running(func):
    @wraps(func)
    def checked_not_running(*args, **kwargs):
        if len(args) == 0 or not isinstance(args[0], Runnable):
            raise ValueError("is_not_running only applicable to Runnable instances")
        runnable = args[0]
        if runnable.running is True:
            raise RunnableAlreadyRunningException(runnable.__class__.__name__)
        return func(*args, **kwargs)

    return checked_not_running


class Runnable(ABC):
    def __init__(self):
        self.__running = False

    @property
    def running(self):
        return self.__running

    def start(self) -> None:
        self._start()
        self.__running = True

    @abstractmethod
    def _start(self) -> None:
        pass

    def stop(self) -> None:
        self._stop()
        self.__running = False

    @abstractmethod
    def _stop(self) -> None:
        pass



from logging import Logger

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.middleware.command_middleware import CommandMiddleware


class LoggingCommandMiddleware(CommandMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger

    def before_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        self.__logger.info(f"Starting handling command {passenger} with {bus_stop.__class__.__name__}")

    def after_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        self.__logger.info(f"Finished handling command {passenger} with {bus_stop.__class__.__name__}")



from typing import Optional, TypeVar

from pypendency.container import Container

from bus_station.shared_terminal.bus_stop import BusStop
from bus_station.shared_terminal.bus_stop_resolver.bus_stop_resolver import BusStopResolver

S = TypeVar("S", bound=BusStop)


class PypendencyBusStopResolver(BusStopResolver[S]):
    def __init__(self, pypendency_container: Container):
        self.__pypendency_container = pypendency_container

    def resolve_from_fqn(self, bus_stop_fqn: str) -> Optional[S]:
        resolved_instance = self.__pypendency_container.get(bus_stop_fqn)
        if resolved_instance is None:
            return None
        if not issubclass(resolved_instance.__class__, BusStop):
            raise TypeError(f"Instance resolved from {bus_stop_fqn} is not a valid BusStop")
        return resolved_instance  # type: ignore



from typing import Type

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.shared_terminal.rpyc_service import RPyCService


class RPyCCommandService(RPyCService[Command, CommandHandler]):
    def passenger_receiver(
        self, bus_stop: CommandHandler, passenger_class: Type[Command], serialized_passenger: str
    ) -> None:
        command = self._passenger_deserializer.deserialize(serialized_passenger, passenger_cls=passenger_class)
        if not isinstance(command, Command):
            raise TypeError("Input serialized command is not a Command")
        self._passenger_receiver.receive(command, bus_stop)



from typing import ClassVar

from rpyc import ThreadedServer

from bus_station.shared_terminal.rpyc_service import RPyCService


class RPyCServer:

    __SELF_HOST_ADDR: ClassVar[str] = "127.0.0.1"

    def __init__(self, rpyc_service: RPyCService, port: int):
        self.__rpyc_service = rpyc_service
        self.__port = port

    def run(self) -> None:
        rpyc_server = ThreadedServer(
            service=self.__rpyc_service,
            hostname=self.__SELF_HOST_ADDR,
            port=self.__port,
        )
        rpyc_server.start()
        rpyc_server.close()



from enum import Enum
from typing import Type

from bus_station.shared_terminal.broker_connection.connection_parameters.rabbitmq_connection_parameters import (
    RabbitMQConnectionParameters,
)


class BrokerConnectionType(Enum):
    RABBITMQ = RabbitMQConnectionParameters

    def connection_parameters_cls(self) -> Type:
        return self.value



from abc import ABCMeta, abstractmethod
from typing import Any, Iterable, Optional, Type, get_type_hints

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer


class EventRegistry(metaclass=ABCMeta):
    def register(self, consumer: EventConsumer, consumer_contact: Any) -> None:
        consumer_event = self.__get_consumer_event(consumer)
        self._register(consumer_event, consumer, consumer_contact)

    def __get_consumer_event(self, consumer: EventConsumer) -> Type[Event]:
        consumer_typing = get_type_hints(consumer.consume)

        if "event" not in consumer_typing:
            raise TypeError(f"Consumer event not found for {consumer.__class__.__name__}")

        if not issubclass(consumer_typing["event"], Event):
            raise TypeError(f"Wrong type for consume event of {consumer.__class__.__name__}")

        return consumer_typing["event"]

    @abstractmethod
    def _register(self, event: Type[Event], consumer: EventConsumer, consumer_contact: Any) -> None:
        pass

    @abstractmethod
    def get_event_destinations(self, event: Type[Event]) -> Optional[Iterable[EventConsumer]]:
        pass

    @abstractmethod
    def get_event_destination_contacts(self, event: Type[Event]) -> Optional[Iterable[Any]]:
        pass

    @abstractmethod
    def get_events_registered(self) -> Iterable[Type[Event]]:
        pass

    @abstractmethod
    def unregister(self, event: Type[Event]) -> None:
        pass



import json

from bus_station.query_terminal.query_response import QueryResponse
from bus_station.query_terminal.serialization.query_response_deserializer import QueryResponseDeserializer


class QueryResponseJSONDeserializer(QueryResponseDeserializer):
    def deserialize(self, query_response_serialized: str) -> QueryResponse:
        query_response_dict = json.loads(query_response_serialized)
        return QueryResponse(data=query_response_dict.get("data"))



from abc import abstractmethod
from typing import Any, Optional, Protocol, TypeVar

from bus_station.passengers.passenger import Passenger

P = TypeVar("P", bound=Passenger)


class Bus(Protocol[P]):
    @abstractmethod
    def transport(self, passenger: P) -> Optional[Any]:
        pass



from abc import abstractmethod
from typing import Protocol

from bus_station.query_terminal.query_response import QueryResponse


class QueryResponseDeserializer(Protocol):
    @abstractmethod
    def deserialize(self, query_response_serialized: str) -> QueryResponse:
        pass



from bus_station.command_terminal.command import Command


class CommandExecutionFailed(Exception):
    def __init__(self, command: Command, reason: str):
        self.command = command
        self.reason = reason
        super().__init__(f"{command} execution has failed due to: {reason}")



from multiprocessing import Process, Queue
from typing import List, Tuple

from bus_station.event_terminal.bus.event_bus import EventBus
from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.registry.in_memory_event_registry import InMemoryEventRegistry
from bus_station.passengers.process_passenger_worker import ProcessPassengerWorker
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer
from bus_station.shared_terminal.runnable import Runnable, is_running


class ProcessEventBus(EventBus, Runnable):
    def __init__(
        self,
        event_serializer: PassengerSerializer,
        event_deserializer: PassengerDeserializer,
        event_registry: InMemoryEventRegistry,
        event_receiver: PassengerReceiver[Event, EventConsumer],
    ):
        EventBus.__init__(self, event_receiver)
        Runnable.__init__(self)
        self.__event_workers: List[ProcessPassengerWorker] = []
        self.__event_worker_processes: List[Process] = []
        self.__event_serializer = event_serializer
        self.__event_deserializer = event_deserializer
        self.__event_registry = event_registry

    def _start(self) -> None:
        for event in self.__event_registry.get_events_registered():
            consumers = self.__event_registry.get_event_destinations(event)
            consumer_queues = self.__event_registry.get_event_destination_contacts(event)
            if consumers is None or consumer_queues is None:
                continue
            for consumer, consumer_queue in zip(consumers, consumer_queues):
                worker, process = self.__create_consumer(consumer, consumer_queue)
                process.start()
                self.__event_workers.append(worker)
                self.__event_worker_processes.append(process)

    def __create_consumer(
        self, consumer: EventConsumer, consumer_queue: Queue
    ) -> Tuple[ProcessPassengerWorker, Process]:
        consumer_worker = ProcessPassengerWorker(
            consumer_queue, consumer, self._event_receiver, self.__event_deserializer
        )
        consumer_process = Process(target=consumer_worker.work)
        return consumer_worker, consumer_process

    @is_running
    def transport(self, passenger: Event) -> None:
        event_consumer_queues = self.__event_registry.get_event_destination_contacts(passenger.__class__)
        if event_consumer_queues is None:
            return

        for event_queue in event_consumer_queues:
            self.__put_event(event_queue, passenger)

    def __put_event(self, queue: Queue, event: Event) -> None:
        serialized_event = self.__event_serializer.serialize(event)
        queue.put(serialized_event)

    def _stop(self) -> None:
        for worker in self.__event_workers:
            worker.stop()
        for process in self.__event_worker_processes:
            process.join()



from abc import abstractmethod
from typing import Optional, Type

from pymongo.database import Database

from bus_station.tracking_terminal.models.passenger_tracking import PassengerTracking
from bus_station.tracking_terminal.repositories.implementations.pymongo.mongo_passenger_tracking_deserializer import (
    MongoPassengerTrackingDeserializer,
)
from bus_station.tracking_terminal.repositories.implementations.pymongo.mongo_passenger_tracking_serializer import (
    MongoPassengerTrackingSerializer,
)
from bus_station.tracking_terminal.repositories.passenger_tracking_repository import PassengerTrackingRepository


class PyMongoPassengerTrackingRepository(PassengerTrackingRepository):
    def __init__(
        self,
        py_mongo_db: Database,
        mongo_tracking_serializer: MongoPassengerTrackingSerializer,
        mongo_tracking_deserializer: MongoPassengerTrackingDeserializer,
    ):
        self.__collection = py_mongo_db.get_collection(self._collection)
        self.__serializer = mongo_tracking_serializer
        self.__deserializer = mongo_tracking_deserializer

    @property
    def model(self) -> Type[PassengerTracking]:
        return PassengerTracking

    @property
    @abstractmethod
    def _collection(self) -> str:
        pass

    def save(self, passenger_tracking: PassengerTracking) -> None:
        serialized_passenger_tracking = self.__serializer.serialize(passenger_tracking)
        self.__collection.update_one(
            filter={"_id": passenger_tracking.id}, update={"$set": serialized_passenger_tracking}, upsert=True
        )

    def find_by_id(self, passenger_tracking_id: str) -> Optional[PassengerTracking]:
        serialized_passenger_tracking = self.__collection.find_one(filter={"_id": passenger_tracking_id})
        if serialized_passenger_tracking is None:
            return None
        return self.__deserializer.deserialize(serialized_passenger_tracking, self.model)



from abc import abstractmethod
from functools import partial
from typing import ClassVar, Generic, Optional, Type, TypeVar

from rpyc import Service

from bus_station.passengers.passenger import Passenger
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.shared_terminal.bus_stop import BusStop

P = TypeVar("P", bound=Passenger)
S = TypeVar("S", bound=BusStop)


class RPyCService(Service, Generic[P, S]):
    __EXPOSED_CALLABLE_PATTERN: ClassVar[str] = "exposed_{callable_name}"

    def __init__(
        self,
        passenger_deserializer: PassengerDeserializer,
        passenger_receiver: PassengerReceiver,
    ):
        self._passenger_deserializer = passenger_deserializer
        self._passenger_receiver = passenger_receiver

    def register(self, passenger_class: Type[P], bus_stop: S) -> None:
        query_callable = partial(self.passenger_receiver, bus_stop, passenger_class)
        exposed_callable_name = self.__EXPOSED_CALLABLE_PATTERN.format(callable_name=passenger_class.__name__)
        setattr(self, exposed_callable_name, query_callable)

    @abstractmethod
    def passenger_receiver(self, bus_stop: S, passenger_class: Type[P], serialized_passenger: str) -> Optional[str]:
        pass



from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.query_terminal.bus.query_bus import QueryBus
from bus_station.query_terminal.handler_not_found_for_query import HandlerNotFoundForQuery
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse
from bus_station.query_terminal.registry.in_memory_query_registry import InMemoryQueryRegistry


class SyncQueryBus(QueryBus):
    def __init__(self, query_registry: InMemoryQueryRegistry, query_receiver: PassengerReceiver[Query, QueryHandler]):
        super().__init__(query_receiver)
        self.__query_registry = query_registry

    def transport(self, passenger: Query) -> QueryResponse:
        query_handler = self.__query_registry.get_query_destination_contact(passenger.__class__)
        if query_handler is None:
            raise HandlerNotFoundForQuery(passenger.__class__.__name__)
        return self._query_receiver.receive(passenger, query_handler)



from typing import Any, Optional, Set, Type

from bus_station.passengers.passenger_class_resolver import PassengerClassResolver
from bus_station.passengers.passenger_record.in_memory_passenger_record_repository import (
    InMemoryPassengerRecordRepository,
)
from bus_station.passengers.passenger_record.passenger_record import PassengerRecord
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.registry.query_registry import QueryRegistry
from bus_station.shared_terminal.bus_stop_resolver.bus_stop_resolver import BusStopResolver
from bus_station.shared_terminal.fqn_getter import FQNGetter


class InMemoryQueryRegistry(QueryRegistry):
    def __init__(
        self,
        in_memory_repository: InMemoryPassengerRecordRepository,
        fqn_getter: FQNGetter,
        query_handler_resolver: BusStopResolver[QueryHandler],
        passenger_class_resolver: PassengerClassResolver,
    ):
        self.__in_memory_repository = in_memory_repository
        self.__fqn_getter = fqn_getter
        self.__query_handler_resolver = query_handler_resolver
        self.__passenger_class_resolver = passenger_class_resolver

    def _register(self, query: Type[Query], handler: QueryHandler, handler_contact: Any) -> None:
        self.__in_memory_repository.save(
            PassengerRecord(
                passenger_name=query.__name__,
                passenger_fqn=self.__fqn_getter.get(query),
                destination_fqn=self.__fqn_getter.get(handler),
                destination_contact=handler_contact,
            )
        )

    def get_query_destination(self, query: Type[Query]) -> Optional[QueryHandler]:
        query_records = self.__in_memory_repository.find_by_passenger_name(query.__name__)
        if query_records is None:
            return None

        query_handler_fqn = query_records[0].destination_fqn
        return self.__query_handler_resolver.resolve_from_fqn(query_handler_fqn)

    def get_query_destination_contact(self, query: Type[Query]) -> Optional[Any]:
        query_records = self.__in_memory_repository.find_by_passenger_name(query.__name__)
        if query_records is None:
            return None
        return query_records[0].destination_contact

    def get_queries_registered(self) -> Set[Type[Query]]:
        queries_registered = set()
        for query_record in self.__in_memory_repository.all():
            query = self.__passenger_class_resolver.resolve_from_fqn(query_record.passenger_fqn)
            if query is None or not issubclass(query, Query):
                continue
            queries_registered.add(query)
        return queries_registered

    def unregister(self, query: Type[Query]) -> None:
        self.__in_memory_repository.delete_by_passenger_name(query.__name__)



from typing import Type

from bus_station.tracking_terminal.models.command_tracking import CommandTracking
from bus_station.tracking_terminal.repositories.command_tracking_repository import CommandTrackingRepository
from bus_station.tracking_terminal.repositories.implementations.pymongo.pymongo_passenger_tracking_repository import (
    PyMongoPassengerTrackingRepository,
)


class PyMongoCommandTrackingRepository(PyMongoPassengerTrackingRepository, CommandTrackingRepository):
    @property
    def model(self) -> Type[CommandTracking]:
        return CommandTracking

    @property
    def _collection(self) -> str:
        return "command_tracking"



import time
from logging import Logger

from bus_station.query_terminal.middleware.query_middleware import QueryMiddleware
from bus_station.query_terminal.query import Query
from bus_station.query_terminal.query_handler import QueryHandler
from bus_station.query_terminal.query_response import QueryResponse


class TimingQueryMiddleware(QueryMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger
        self.__start_time = None

    def before_handle(self, passenger: Query, bus_stop: QueryHandler) -> None:
        self.__start_time = time.time()

    def after_handle(self, passenger: Query, bus_stop: QueryHandler, query_response: QueryResponse) -> QueryResponse:
        execution_time = time.time() - self.__start_time
        self.__logger.info(f"Query {passenger} handled by {bus_stop.__class__.__name__} in {execution_time} seconds")
        return query_response



from multiprocessing import Process, Queue
from typing import List, Tuple

from bus_station.command_terminal.bus.command_bus import CommandBus
from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.in_memory_command_registry import InMemoryCommandRegistry
from bus_station.passengers.process_passenger_worker import ProcessPassengerWorker
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer
from bus_station.shared_terminal.runnable import Runnable, is_running


class ProcessCommandBus(CommandBus, Runnable):
    def __init__(
        self,
        command_serializer: PassengerSerializer,
        command_deserializer: PassengerDeserializer,
        command_registry: InMemoryCommandRegistry,
        command_receiver: PassengerReceiver[Command, CommandHandler],
    ):
        CommandBus.__init__(self, command_receiver)
        Runnable.__init__(self)
        self.__command_workers: List[ProcessPassengerWorker] = []
        self.__command_worker_processes: List[Process] = []
        self.__command_serializer = command_serializer
        self.__command_deserializer = command_deserializer
        self.__command_registry = command_registry

    def _start(self) -> None:
        for command in self.__command_registry.get_commands_registered():
            handler = self.__command_registry.get_command_destination(command)
            handler_queue = self.__command_registry.get_command_destination_contact(command)
            if handler_queue is None or handler is None:
                continue
            worker, process = self.__create_handler(handler, handler_queue)
            self.__command_workers.append(worker)
            self.__command_worker_processes.append(process)
            process.start()

    def __create_handler(
        self, command_handler: CommandHandler, command_queue: Queue
    ) -> Tuple[ProcessPassengerWorker, Process]:
        handler_worker = ProcessPassengerWorker(
            command_queue, command_handler, self._command_receiver, self.__command_deserializer
        )
        handler_process = Process(target=handler_worker.work)
        return handler_worker, handler_process

    @is_running
    def transport(self, passenger: Command) -> None:
        handler_queue = self.__command_registry.get_command_destination_contact(passenger.__class__)
        if handler_queue is None:
            raise HandlerNotFoundForCommand(passenger.__class__.__name__)

        self.__put_command(handler_queue, passenger)

    def __put_command(self, queue: Queue, command: Command) -> None:
        serialized_command = self.__command_serializer.serialize(command)
        queue.put(serialized_command)

    def _stop(self) -> None:
        for worker in self.__command_workers:
            worker.stop()
        for process in self.__command_worker_processes:
            process.join()



from kombu.connection import Connection

from bus_station.shared_terminal.broker_connection.connection_parameters.connection_parameters import (
    ConnectionParameters,
)


class KombuConnectionFactory:
    def get_connection(self, broker_connection_parameters: ConnectionParameters) -> Connection:
        return Connection(broker_connection_parameters.get_connection_string())



from abc import abstractmethod

from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.shared_terminal.bus import Bus


class EventBus(Bus[Event]):
    def __init__(self, event_receiver: PassengerReceiver[Event, EventConsumer]):
        self._event_receiver = event_receiver

    @abstractmethod
    def transport(self, passenger: Event) -> None:
        pass



from multiprocessing import Process
from typing import ClassVar, List, Optional, Tuple, Type

from kombu import Connection
from kombu.messaging import Exchange, Producer, Queue
from kombu.transport.virtual import Channel

from bus_station.command_terminal.bus.command_bus import CommandBus
from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.remote_command_registry import RemoteCommandRegistry
from bus_station.passengers.passenger_kombu_consumer import PassengerKombuConsumer
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer
from bus_station.shared_terminal.runnable import Runnable


class KombuCommandBus(CommandBus, Runnable):
    __DEAD_LETTER_EXCHANGE_NAME: ClassVar[str] = "failed_commands"

    def __init__(
        self,
        broker_connection: Connection,
        command_serializer: PassengerSerializer,
        command_deserializer: PassengerDeserializer,
        command_registry: RemoteCommandRegistry,
        command_receiver: PassengerReceiver[Command, CommandHandler],
    ):
        CommandBus.__init__(self, command_receiver)
        Runnable.__init__(self)
        self.__broker_connection = broker_connection
        self.__command_consumers: List[PassengerKombuConsumer] = list()
        self.__command_consumer_processes: List[Process] = list()
        self.__command_serializer = command_serializer
        self.__command_deserializer = command_deserializer
        self.__command_registry = command_registry
        self.__producer: Optional[Producer] = None

    def _start(self) -> None:
        broker_channel = self.__broker_connection.channel()
        self.__create_dead_letter_exchange(broker_channel)

        for command in self.__command_registry.get_commands_registered():
            handler = self.__command_registry.get_command_destination(command)
            handler_queue_name = self.__command_registry.get_command_destination_contact(command)
            if handler_queue_name is None or handler is None:
                continue
            consumer, consumer_process, consumer_queue = self.__create_consumer(handler, handler_queue_name, command)
            self.__command_consumers.append(consumer)
            self.__command_consumer_processes.append(consumer_process)
            consumer_queue.declare(channel=broker_channel)
            consumer_process.start()

        self.__producer = Producer(broker_channel)

    def __create_dead_letter_exchange(self, channel: Channel) -> None:
        command_failure_exchange = Exchange(self.__DEAD_LETTER_EXCHANGE_NAME, type="fanout", channel=channel)
        command_failure_exchange.declare()

    def __create_consumer(
        self, command_handler: CommandHandler, handler_queue_name: str, command_cls: Type[Command]
    ) -> Tuple[PassengerKombuConsumer, Process, Queue]:
        handler_queue = Queue(
            handler_queue_name, queue_arguments={"x-dead-letter-exchange": self.__DEAD_LETTER_EXCHANGE_NAME}
        )
        handler_consumer = PassengerKombuConsumer(
            self.__broker_connection,
            handler_queue,
            command_handler,
            command_cls,
            self._command_receiver,
            self.__command_deserializer,
        )
        handler_process = Process(target=handler_consumer.run)
        return handler_consumer, handler_process, handler_queue

    def transport(self, passenger: Command) -> None:
        handler_queue_name = self.__command_registry.get_command_destination_contact(passenger.__class__)
        if handler_queue_name is None:
            raise HandlerNotFoundForCommand(passenger.__class__.__name__)

        self.__publish_command(passenger, handler_queue_name)

    def __publish_command(self, command: Command, routing_key: str) -> None:
        serialized_command = self.__command_serializer.serialize(command)
        if self.__producer is not None:
            self.__producer.publish(
                serialized_command,
                exchange="",
                routing_key=routing_key,
                retry=True,
                retry_policy={
                    "interval_start": 0,
                    "interval_step": 2,
                    "interval_max": 10,
                    "max_retries": 10,
                },
            )

    def _stop(self) -> None:
        for consumer in self.__command_consumers:
            consumer.stop()
        for process in self.__command_consumer_processes:
            process.join()
        if self.__producer is not None:
            self.__producer.release()
        self.__broker_connection.release()



from multiprocessing import Process
from typing import ClassVar, List, Optional, Tuple, Type

from kombu import Connection
from kombu.messaging import Exchange, Producer, Queue
from kombu.transport.virtual import Channel

from bus_station.event_terminal.bus.event_bus import EventBus
from bus_station.event_terminal.event import Event
from bus_station.event_terminal.event_consumer import EventConsumer
from bus_station.event_terminal.registry.remote_event_registry import RemoteEventRegistry
from bus_station.passengers.passenger import Passenger
from bus_station.passengers.passenger_kombu_consumer import PassengerKombuConsumer
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer
from bus_station.shared_terminal.bus_stop import BusStop
from bus_station.shared_terminal.runnable import Runnable


class KombuEventBus(EventBus, Runnable):
    __DEAD_LETTER_EXCHANGE_NAME: ClassVar[str] = "failed_events"

    def __init__(
        self,
        broker_connection: Connection,
        event_serializer: PassengerSerializer,
        event_deserializer: PassengerDeserializer,
        event_registry: RemoteEventRegistry,
        event_receiver: PassengerReceiver[Event, EventConsumer],
    ):
        EventBus.__init__(self, event_receiver)
        Runnable.__init__(self)
        self.__broker_connection = broker_connection
        self.__event_kombu_consumers: List[PassengerKombuConsumer] = []
        self.__event_kombu_consumer_processes: List[Process] = []
        self.__event_serializer = event_serializer
        self.__event_deserializer = event_deserializer
        self.__event_registry = event_registry
        self.__producer: Optional[Producer] = None

    def _start(self) -> None:
        broker_channel = self.__broker_connection.channel()
        self.__create_dead_letter_exchange(broker_channel)

        for event in self.__event_registry.get_events_registered():
            event_consumers = self.__event_registry.get_event_destinations(event)
            exchange_names = self.__event_registry.get_event_destination_contacts(event)
            if event_consumers is None or exchange_names is None:
                continue
            for event_consumer, event_exchange_name in zip(event_consumers, exchange_names):
                event_exchange = self.__create_event_exchange(event_exchange_name)
                event_exchange.declare(channel=broker_channel)
                kombu_consumer, kombu_consumer_process, kombu_queue = self.__create_consumer(
                    event_consumer, event, event_exchange
                )
                kombu_queue.declare(channel=broker_channel)
                kombu_consumer_process.start()
                self.__event_kombu_consumers.append(kombu_consumer)
                self.__event_kombu_consumer_processes.append(kombu_consumer_process)

        self.__producer = Producer(broker_channel)

    def __create_dead_letter_exchange(self, channel: Channel) -> None:
        event_failure_exchange = Exchange(self.__DEAD_LETTER_EXCHANGE_NAME, type="fanout", channel=channel)
        event_failure_exchange.declare()

    def __create_event_exchange(self, event_name: str) -> Exchange:
        event_exchange = Exchange(event_name, type="fanout")
        return event_exchange

    def __create_consumer(
        self, event_consumer: BusStop, event_cls: Type[Passenger], exchange: Exchange
    ) -> Tuple[PassengerKombuConsumer, Process, Queue]:
        consumer_queue = Queue(
            event_consumer.__class__.__name__,
            exchange=exchange,
            queue_arguments={"x-dead-letter-exchange": self.__DEAD_LETTER_EXCHANGE_NAME},
        )
        consumer_consumer = PassengerKombuConsumer(
            self.__broker_connection,
            consumer_queue,
            event_consumer,
            event_cls,
            self._event_receiver,
            self.__event_deserializer,
        )
        consumer_process = Process(target=consumer_consumer.run)
        return consumer_consumer, consumer_process, consumer_queue

    def transport(self, passenger: Event) -> None:
        event_exchange_names = self.__event_registry.get_event_destination_contacts(passenger.__class__)
        if event_exchange_names is None:
            return

        published_exchanges = []
        for event_exchange_name in event_exchange_names:
            if event_exchange_name in published_exchanges:
                continue
            self.__publish_event(passenger, event_exchange_name)
            published_exchanges.append(event_exchange_name)

    def __publish_event(self, event: Event, event_exchange_name: str) -> None:
        serialized_event = self.__event_serializer.serialize(event)
        if self.__producer is not None:
            self.__producer.publish(
                serialized_event,
                exchange=event_exchange_name,
                retry=True,
                retry_policy={
                    "interval_start": 0,
                    "interval_step": 2,
                    "interval_max": 10,
                    "max_retries": 10,
                },
            )

    def _stop(self) -> None:
        for kombu_consumer in self.__event_kombu_consumers:
            kombu_consumer.stop()
        for kombu_consumer_process in self.__event_kombu_consumer_processes:
            kombu_consumer_process.join()
        if self.__producer is not None:
            self.__producer.release()
        self.__broker_connection.release()



from abc import abstractmethod
from typing import Protocol

from bus_station.passengers.passenger import Passenger


class PassengerSerializer(Protocol):
    @abstractmethod
    def serialize(self, passenger: Passenger) -> str:
        pass



import time
from logging import Logger

from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.middleware.command_middleware import CommandMiddleware


class TimingCommandMiddleware(CommandMiddleware):
    def __init__(self, logger: Logger):
        self.__logger = logger
        self.__start_time = None

    def before_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        self.__start_time = time.time()

    def after_handle(self, passenger: Command, bus_stop: CommandHandler) -> None:
        execution_time = time.time() - self.__start_time
        self.__logger.info(f"Command {passenger} handled by {bus_stop.__class__.__name__} in {execution_time} seconds")



class RunnableAlreadyRunningException(Exception):
    def __init__(self, runnable_name: str):
        self.runnable_name = runnable_name
        super().__init__(f"{runnable_name} is already running")



from abc import ABC
from typing import Any, Generic, Iterable, List, Tuple, Type, TypeVar

from bus_station.passengers.passenger import Passenger
from bus_station.passengers.passenger_middleware import PassengerMiddleware
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.shared_terminal.bus_stop import BusStop

M = TypeVar("M", bound=PassengerMiddleware)
S = TypeVar("S", bound=BusStop)
P = TypeVar("P", bound=Passenger)


class PassengerMiddlewareReceiver(PassengerReceiver[P, S], Generic[P, S, M], ABC):
    def __init__(self):
        self._middleware_definitions: List[M | Tuple[Type[M], Tuple[Any]]] = list()

    def add_middleware_definition(
        self, passenger_middleware_cls: Type[M], *middleware_args: Any, lazy: bool = False
    ) -> None:
        if lazy is False:
            self._middleware_definitions.append(passenger_middleware_cls(*middleware_args))
        else:
            self._middleware_definitions.append((passenger_middleware_cls, middleware_args))

    def _get_middlewares(self) -> Iterable[M]:
        for middleware_definition in self._middleware_definitions:
            if isinstance(middleware_definition, tuple):
                middleware_cls, middleware_args = middleware_definition
                yield middleware_cls(*middleware_args)  # pyre-ignore[19]
            else:
                yield middleware_definition



import os
import signal
from multiprocessing.context import Process
from typing import ClassVar, Optional

from rpyc import Connection, connect

from bus_station.command_terminal.bus.command_bus import CommandBus
from bus_station.command_terminal.command import Command
from bus_station.command_terminal.command_handler import CommandHandler
from bus_station.command_terminal.handler_not_found_for_command import HandlerNotFoundForCommand
from bus_station.command_terminal.registry.remote_command_registry import RemoteCommandRegistry
from bus_station.command_terminal.rpyc_command_service import RPyCCommandService
from bus_station.passengers.reception.passenger_receiver import PassengerReceiver
from bus_station.passengers.serialization.passenger_deserializer import PassengerDeserializer
from bus_station.passengers.serialization.passenger_serializer import PassengerSerializer
from bus_station.shared_terminal.rpyc_server import RPyCServer
from bus_station.shared_terminal.runnable import Runnable


class RPyCCommandBus(CommandBus, Runnable):
    __SELF_ADDR_PATTERN: ClassVar[str] = "{host}:{port}"

    def __init__(
        self,
        self_host: str,
        self_port: int,
        command_serializer: PassengerSerializer,
        command_deserializer: PassengerDeserializer,
        command_registry: RemoteCommandRegistry,
        command_receiver: PassengerReceiver[Command, CommandHandler],
    ):
        CommandBus.__init__(self, command_receiver)
        Runnable.__init__(self)
        self.__self_host = self_host
        self.__self_port = self_port
        self.__command_serializer = command_serializer
        self.__command_deserializer = command_deserializer
        self.__command_registry = command_registry
        self.__rpyc_service = RPyCCommandService(self.__command_deserializer, self._command_receiver)
        self.__rpyc_server: Optional[RPyCServer] = None
        self.__server_process: Optional[Process] = None

    def _start(self):
        for command in self.__command_registry.get_commands_registered():
            handler = self.__command_registry.get_command_destination(command)
            self.__rpyc_service.register(command, handler)

        self.__rpyc_server = RPyCServer(
            rpyc_service=self.__rpyc_service,
            port=self.__self_port,
        )
        self.__server_process = Process(target=self.__rpyc_server.run)
        self.__server_process.start()

    def transport(self, passenger: Command) -> None:
        handler_address = self.__command_registry.get_command_destination_contact(passenger.__class__)
        if handler_address is None:
            raise HandlerNotFoundForCommand(passenger.__class__.__name__)

        rpyc_client = self.__get_rpyc_client(handler_address)
        self.__execute_command(rpyc_client, passenger)
        rpyc_client.close()

    def __get_rpyc_client(self, handler_addr: str) -> Connection:
        host, port = handler_addr.split(":")
        return connect(host, port=port)

    def __execute_command(self, client: Connection, command: Command) -> None:
        serialized_command = self.__command_serializer.serialize(command)
        getattr(client.root, command.__class__.__name__)(serialized_command)

    def _stop(self) -> None:
        server_process = self.__server_process
        if server_process is not None and server_process.pid is not None:
            os.kill(server_process.pid, signal.SIGINT)
            server_process.join()
